from enum import Enum


class ConfigSource(str, Enum):
    """
    Possible kinds of configuration sources.
    """
    init = 'init'
    env_var = 'env_var'
    map = 'map'
    file = 'file'
    class_default = 'class_default'

    def __repr__(self):
        return "'" + self.value + "'"

    def __str__(self):
        return str(self.value)



from pathlib import Path
from typing import Optional, Any, Dict, List, Union

from pydantic import BaseModel

from confident.config_source import ConfigSource
from confident.map_field import MAP_FIELD_FLAG

IGNORE_MISSING_FILES_DEFAULT = True
DEFAULT_SOURCE_PRIORITY = [
    ConfigSource.init, ConfigSource.env_var, ConfigSource.map, ConfigSource.file, ConfigSource.class_default
]


class ConfigSpecs(BaseModel):
    """
    A model that holds all the metadata regarding the Confident config object.
    """
    specs_path: Optional[Path] = None
    files: List[Path] = []
    ignore_missing_files: bool = IGNORE_MISSING_FILES_DEFAULT
    map_name: Optional[str] = None
    map_field: Optional[str] = None
    config_map: Optional[Union[Path, Dict[str, Union[Path, Dict[str, Any]]]]] = None
    class_path: Optional[Path] = None
    creation_path: Optional[Path] = None
    source_priority: List[ConfigSource] = DEFAULT_SOURCE_PRIORITY

    @classmethod
    def from_path(cls, path: Union[Path, str], class_path: Optional[Path] = None, creation_path: Optional[Path] = None):
        obj = cls.parse_file(path)
        obj.specs_path = Path(path)
        obj.class_path = class_path or obj.class_path
        obj.creation_path = creation_path or obj.creation_path
        return obj

    @classmethod
    def from_model(cls, model, values, class_path=None, creation_path=None):
        model_config = model.Config.__dict__

        map_field = cls._find_map_field(
            model=model,
            explicit_map_field=values.pop('_map_field', None) or model_config.get('map_field')
        )

        files = values.pop('_files', None) or model_config.get('files')
        files = [files] if isinstance(files, str) else files or []

        ignore_missing_files = values.pop('_ignore_missing_files', None)
        ignore_missing_files = (
            ignore_missing_files if ignore_missing_files is not None
            else model_config.get('ignore_missing_files', IGNORE_MISSING_FILES_DEFAULT)
        )

        obj = cls(
            specs_path=values.pop('_specs_path', None),
            files=files,
            ignore_missing_files=ignore_missing_files,
            map_name=values.pop('_map_name', None) or model_config.get('map_name'),
            map_field=map_field,
            config_map=values.pop('_config_map', None) or model_config.get('config_map'),
            class_path=class_path,
            creation_path=creation_path,
            source_priority=(
                    values.pop('_source_priority', None) or model_config.get('source_priority') or
                    DEFAULT_SOURCE_PRIORITY
            )
        )
        return obj

    @staticmethod
    def _find_map_field(model, explicit_map_field) -> str:
        """
        Searches if one of the fields declared in the subclass has marked as the deployment field.
        Args:
            explicit_map_field: The deployment field received as argument.

        Returns:
            A single deployment field. None if no field provided in any way.

        Raises:
            ValueError - If more than one deployment fields is received.
        """
        properties_marked_as_map_field = [
            name for name, model_field in model.__fields__.items() if
            model_field.field_info.extra.get(MAP_FIELD_FLAG)
        ]
        if not properties_marked_as_map_field:
            return explicit_map_field
        if explicit_map_field and properties_marked_as_map_field:
            raise ValueError(
                f'Cannot have both explicit `_map_field` and also `MapField()` '
                f'in {model.__class__.__name__} declaration'
            )
        if len(properties_marked_as_map_field) > 1:
            raise ValueError(f'Cannot have more then one `MapField()` in {model.__class__.__name__} declaration')
        return properties_marked_as_map_field[0]



from abc import ABC, abstractmethod
from typing import List

from pydantic import BaseSettings

from confident.specs import ConfigSpecs
from confident.config_field import ConfigField


class SourceLoader(ABC):
    NAME = ...

    def __init__(self, specs: ConfigSpecs = None):
        self.specs = specs or ConfigSpecs()

    @abstractmethod
    def load_fields(self, settings: BaseSettings) -> List[ConfigField]:
        ...



from typing import List

from pydantic import BaseSettings

from confident.config_field import ConfigField
from confident.config_source import ConfigSource
from confident.loaders.source_loader_base import SourceLoader


class DefaultSourceLoader(SourceLoader):
    NAME = ConfigSource.class_default

    def load_fields(self, settings: BaseSettings) -> List[ConfigField]:
        """
        Loads default values declared in the inheriting class into a dictionary.
        """
        fields = []
        for field_name, model_field in settings.__fields__.items():
            # If the field is required, there is no default value to load.
            if model_field.required:
                continue
            # Uses `pydantic` ModelField to retrieve the default values of the model.
            default_value = model_field.get_default()
            fields.append(ConfigField(
                name=field_name,
                value=default_value,
                source_name=settings.__class__.__name__,
                source_type=ConfigSource.class_default,
                source_location=self.specs.class_path,
            ))
        return fields



import inspect
from copy import deepcopy
from pathlib import Path
from typing import Tuple, Any, Dict, List

from pydantic import BaseSettings
from pydantic.env_settings import SettingsSourceCallable

from confident.config_field import ConfigField
from confident.config_source import ConfigSource
from confident.loader_manager import LoaderManager
from confident.loaders.default_source_loader import DefaultSourceLoader
from confident.loaders.env_source_loader import EnvSourceLoader
from confident.loaders.file_source_loader import FileSourceLoader
from confident.loaders.init_source_loader import InitSourceLoader
from confident.loaders.map_source_loader import MapSourceLoader
from confident.specs import ConfigSpecs
from confident.utils import get_class_file_path

SPECS_ATTR = '_specs'
LOADER_MANAGER_ATTR = '_loader_manager'


class BaseConfig(BaseSettings):
    __slots__ = (SPECS_ATTR, LOADER_MANAGER_ATTR)

    def __init__(self, **values: Any):
        # Prepare metadata.
        subclass_location = get_class_file_path(cls=self)
        caller_module = inspect.getmodule(inspect.stack()[1][0])
        caller_location = caller_module.__file__ if caller_module else Path.cwd()

        specs = self.Config.__dict__.get('specs')
        if not specs:
            specs_path = values.pop('_specs_path', None) or self.Config.__dict__.get('specs_path')
            if specs_path:
                specs = ConfigSpecs.from_path(
                    path=specs_path, class_path=subclass_location, creation_path=caller_location
                )
            else:
                specs = ConfigSpecs.from_model(
                    model=self, values=values, class_path=subclass_location, creation_path=caller_location
                )

        loader_manager = LoaderManager(settings_obj=self, source_priority=specs.source_priority)
        object.__setattr__(self, SPECS_ATTR, specs)
        object.__setattr__(self, LOADER_MANAGER_ATTR, loader_manager)

        # Create temporary context for the scope of `Config` class.
        self.Config._confident_loader_manager_context_ = loader_manager
        self.Config._confident_specs_context_ = specs

        super().__init__(**values)

        # Reset the context of the Config class, so it will be clean for the next object creation.
        del self.Config._confident_loader_manager_context_
        del self.Config._confident_specs_context_

    def _build_values(self, init_kwargs: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        values_dict = super()._build_values(init_kwargs, **kwargs)
        loader_manager = self.__getattribute__(LOADER_MANAGER_ATTR)
        loader_manager.full_fields = values_dict

        return {name: field.value for name, field in values_dict.items()}

    class Config:
        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            # Get BaseSettings default loaders callables.
            loader_manager = cls._confident_loader_manager_context_
            specs = cls._confident_specs_context_
            loader_manager.init_settings_callable = init_settings
            loader_manager.env_settings_callable = env_settings
            loader_manager.file_secret_settings_callable = file_secret_settings

            # Add all the default loaders.
            loader_manager.loaders.extend([
                InitSourceLoader(specs=specs, init_settings_callable=init_settings),
                EnvSourceLoader(specs=specs, env_settings_callable=env_settings),
                MapSourceLoader(specs=specs, all_loaded_fields=loader_manager.all_loaded_fields),
                FileSourceLoader(specs=specs),
                DefaultSourceLoader(specs=specs)
            ])

            return loader_manager.load_all()

    @property
    def __specs__(self) -> ConfigSpecs:
        """
        Returns: The model specs.
        """
        return self.__getattribute__(SPECS_ATTR)

    def specs(self):
        return deepcopy(self.__specs__)

    @property
    def __source_priority__(self) -> List[ConfigSource]:
        """
        Returns: List of sources from the highest priority to the lowest.
        """
        return self.__specs__.source_priority

    def source_priority(self):
        return deepcopy(self.__source_priority__)

    @property
    def __full_fields__(self) -> Dict[str, Any]:
        """
        Returns: A dictionary with details of every field.
        """
        return self.__getattribute__(LOADER_MANAGER_ATTR).full_fields

    def full_fields(self):
        return deepcopy(self.__full_fields__)

    @property
    def __all_loaded_fields__(self) -> Dict[str, List[ConfigField]]:
        """
        Returns: A dictionary with all the fields that were loaded before the prioritization classified by sources.
        """
        return self.__getattribute__(LOADER_MANAGER_ATTR).all_loaded_fields

    def all_loaded_fields(self):
        return deepcopy(self.__all_loaded_fields__)


class Confident(BaseConfig):
    pass



import os
from typing import List

from pydantic import BaseSettings

from confident.config_field import ConfigField
from confident.config_source import ConfigSource
from confident.loaders.source_loader_base import SourceLoader
from confident.utils import load_file, convert_field_value


class FileSourceLoader(SourceLoader):
    NAME = ConfigSource.file

    def load_fields(self, settings: BaseSettings) -> List[ConfigField]:
        """
        Finds and loads requested config fields from files into a dictionary.

        Raises:
            ValueError -
                If file is not exists and ignore_missing_files=False.
                If the file is not in a supported format.
        """
        fields = {}
        for file_path in self.specs.files:
            if not os.path.isfile(file_path) and self.specs.ignore_missing_files:
                continue
            file_dict = load_file(path=file_path)

            # Creates a dict with `ConfigField` from the file data and merges them with the rest of the properties
            # from other files.
            fields.update(
                {key: ConfigField(
                    name=key,
                    value=convert_field_value(settings=settings, field_name=key, origin_value=value),
                    origin_value=value,
                    source_name=os.path.basename(file_path),
                    source_type=ConfigSource.file,
                    source_location=file_path
                ) for key, value in file_dict.items()})

        # Returns only the properties that are relevant to the class definition.
        return [fields[key] for key in fields.keys() & settings.__fields__.keys()]




import importlib
import json
from pathlib import Path
from typing import Any, Dict, Union

import yaml
from pydantic import BaseSettings


def load_file(path: Union[Path, str]) -> Dict[str, Any]:
    """
    Loads fields from a file into a dictionary.

    Args:
        path: Path to the file to load.

    Raises:
        ValueError - If the file is not exists.
        ValueError - If the file format is not supported.
        ValueError - If the loaded data is not a dict.
    """
    path = Path(path)

    if not path.is_file():
        raise ValueError(f'{path=} is not exists.')

    if path.suffix == '.json':
        with open(path, mode='r') as file:
            loaded = json.load(file)

    elif path.suffix == '.yaml':
        with open(path, mode='r') as file:
            loaded = yaml.safe_load(file)

    else:
        raise ValueError(f'{path=} is not a supported file.')

    # Check the loaded data
    if loaded is None:
        loaded = {}
    if not isinstance(loaded, dict):
        raise ValueError(f'{path=} has to have a valid dict content.')

    return loaded


def get_class_file_path(cls: object) -> Union[str, Path]:
    """
    Gets the path that the config class is initiated from.
    Tries to find the path of the caller module.
    If there is no module that called this class constructor, returns the current working path.
    Mainly happens when running on terminal.
    """
    try:
        return importlib.import_module(cls.__module__).__file__
    except (AttributeError, ImportError):
        return Path.cwd()


def convert_field_value(settings: BaseSettings, field_name: str, origin_value: Any) -> Any:
    """
    Tries to convert the field value to the right type in the declaration annotation.
    If nothing works, do nothing and let pydantic `BaseSettings` `__init__` call handle the validation.
    Args:
        settings: The BaseSetting object with all config fields to be loaded.
        field_name: The name of the attribute to find its expected type.
        origin_value: The original value retrieved from the config source.

    Returns:
        The converted origin value. Can also be untouched.
    """
    model_field = settings.__fields__.get(field_name)
    # model_field.outer_type_ can be type annotation and not type.
    if (model_field and isinstance(model_field.outer_type_, type)
            and isinstance(origin_value, model_field.outer_type_)):
        return origin_value
    if isinstance(origin_value, str):
        try:
            return json.loads(origin_value)
        except (TypeError, ValueError):
            pass
    return origin_value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['confident', 'confident.loaders']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.0,<2.0.0', 'pyyaml>=6.0,<7.0']

setup_kwargs = {
    'name': 'confident',
    'version': '0.2.6',
    'description': 'Loading configurations from multiple sources into a data model.',
    'long_description': '# Confident\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/confident?style=plastic)](https://github.com/limonyellow/confident)\n[![PyPI](https://img.shields.io/pypi/v/confident?style=plastic&color=%2334D058)](https://pypi.org/project/confident/)\n[![GitHub Workflow Status (branch)](https://img.shields.io/github/workflow/status/limonyellow/confident/Python%20package/main?style=plastic)](https://github.com/limonyellow/confident/actions)\n[![GitHub](https://img.shields.io/github/license/limonyellow/confident?style=plastic)](https://github.com/limonyellow/confident)\n[![Coverage](https://codecov.io/gh/limonyellow/confident/branch/main/graph/badge.svg?token=L161UYE2RM)](https://codecov.io/gh/limonyellow/confident)\n---\n\n[**Documentation**](https://limonyellow.github.io/confident/)\n\n---\n\nConfident helps you create configuration objects from multiple sources such as files, environment variables and maps.  \nConfident BaseConfig class is a data model that enforce validation and type hints by using [pydantic](https://pydantic-docs.helpmanual.io/) library.\n\nWith Confident you can manage multiple configurations depends on the environment your code is deployed.\nWhile having lots of flexibility how to describe your config objects, Confident will provide visibility of the process \nand help you expose misconfiguration as soon as possible.\n\n\n## Example\n\n```python\nimport os\n\nfrom confident import BaseConfig\n\n\n# Creating your own config class by inheriting from `BaseConfig`.\nclass MyAppConfig(BaseConfig):\n    port: int = 5000\n    host: str = \'localhost\'\n    labels: list\n\n\n# Illustrates some environment variables.\nos.environ[\'host\'] = \'127.0.0.1\'\nos.environ[\'labels\'] = \'["FOO", "BAR"]\'  # JSON strings can be used for more types.\n\n# Creating the config object. `BaseConfig` will load the values of the properties.\nconfig = MyAppConfig()\n\nprint(config.host)\n# > 127.0.0.1\nprint(config.json())\n# > {"port": 5000, "host": "127.0.0.1", "labels": ["FOO", "BAR"]}\nprint(config)\n# > port=5000 host=\'127.0.0.1\' labels=[\'FOO\', \'BAR\']\nprint(config.full_fields())\n# > {\n# \'port\': ConfigField(name=\'port\', value=5000, origin_value=5000, source_name=\'MyAppConfig\', source_type=\'class_default\', source_location=PosixPath(\'~/confident/readme_example.py\')),\n# \'host\': ConfigField(name=\'host\', value=\'127.0.0.1\', origin_value=\'127.0.0.1\', source_name=\'host\', source_type=\'env_var\', source_location=\'host\'),\n# \'labels\': ConfigField(name=\'labels\', value=[\'FOO\', \'BAR\'], origin_value=\'["FOO", "BAR"]\', source_name=\'labels\', source_type=\'env_var\', source_location=\'labels\')\n# }\n\n```\n\n## Installation\n```shell\n(.venv) $ pip install confident\n```\n\n## Capabilities\n### Customized Fields Loaders\nBuilt-in loaders:  \n- Environment variables.  \n- Config files such as \'json\' and \'yaml\'.  \n- Config maps to load fields depends on the environment. (See documentation)  \n\nIt is possible to configure the loading priority and add your own loader classes.\n\n### Full Support of Pydantic BaseSettings\nConfident core functionality is based on [pydantic](https://pydantic-docs.helpmanual.io/) library. \nThat means `BaseConfig` object has all the benefits of pydantic\'s [`BaseModel`](https://pydantic-docs.helpmanual.io/usage/models/) \nand [`BaseSettings`](https://pydantic-docs.helpmanual.io/usage/settings/)\nincluding type validation, [object transformation](https://pydantic-docs.helpmanual.io/usage/exporting_models/) and many more features.\n\n### Config Loading visibility\n`BaseConfig` object stores details about the fields loading process and offers ways to understand the source of each loaded field.\nDetails about the origin value (before conversion), the location of the source and the type of loader, can all be accessed from the object. \n\n## Examples\nMore examples can be found in the project\'s [repository](https://github.com/limonyellow/confident).\n\n## Contributing\nTo contribute to Confident, please make sure any new features or changes to existing functionality include test coverage.\n',
    'author': 'limonyellow',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/limonyellow/confident',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import List

from pydantic import BaseSettings

from confident.config_field import ConfigField
from confident.config_source import ConfigSource
from confident.loaders.source_loader_base import SourceLoader
from confident.utils import load_file, convert_field_value


class MapSourceLoader(SourceLoader):
    NAME = ConfigSource.map

    def __init__(self, all_loaded_fields: dict, **kwargs):
        super().__init__(**kwargs)
        self.all_loaded_fields = all_loaded_fields

    def load_fields(self, settings: BaseSettings) -> List[ConfigField]:
        """
        Loads the relevant map config properties.

        Raises:
            ValueError - If wrong combination or values of fields is detected in the following cases:
                If `map_field` and also `map_name` are present.
                If no `config_map` is provided.
                If the `map_name` is not of type `str`.
                If the `map_field` appears inside the `config_map`.
        """
        # Extracts the map metadata field and validates them.
        map_name = self.specs.map_name
        map_field = self.specs.map_field
        config_map = self.specs.config_map
        map_location = self.specs.creation_path

        if map_field is None and map_name is None:
            return []
        if map_field is not None and map_name is not None:
            raise ValueError('Cannot have both `map_field` and `map_name`. Only one can be used.')
        if config_map is None:
            raise ValueError('No `config_map` was provided.')
        if isinstance(config_map, Path):
            map_location = config_map
            config_map = load_file(map_location)

        selected_config = {}
        config_fields = []

        # According to the map name, extracts the chosen config.
        if map_name:
            selected_config = config_map.get(map_name)
        if map_field:
            # Search for the map name in all possible sources ordered by priority.
            for source in self.specs.source_priority:
                if source == ConfigSource.map:
                    continue
                config_property = self.all_loaded_fields[source].get(map_field)
                if config_property:
                    map_name = config_property.value
                    break

            if not isinstance(map_name, str):
                raise ValueError(
                    f'{map_field=} is not valid. Value has to be <str> not "{map_name}" '
                    f'type={type(map_name)}'
                )
            selected_config = config_map.get(map_name)

        if selected_config is None:
            raise KeyError(f'No matching map config to {map_name=}. Check your `config_map`.')

        if isinstance(selected_config, str) or isinstance(selected_config, Path):
            selected_config = load_file(selected_config)

        # Creates the `ConfigField` dictionary.
        for name, value in selected_config.items():
            if name == map_field:
                raise ValueError(
                    f"{map_field=} cannot appear in the map config key '{map_name}'. "
                    f"Look for {map_location=} at '{map_name}'. "
                    f"Remove '{map_field}' key or change the map field."
                )
            config_fields.append(
                ConfigField(
                    name=name,
                    value=convert_field_value(settings=settings, field_name=name, origin_value=value),
                    origin_value=value,
                    source_name=map_name,
                    source_type=ConfigSource.map,
                    source_location=map_location,
                )
            )

        self.specs.map_name = map_name
        return config_fields
